/*
 * ${license.title}
 * Copyright (C) 2022 ${license.owner}
 * ${license.mailto}
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 3 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 */

#include "fggl/phys/bullet/types.hpp"

namespace fggl::phys::bullet {

	BulletPhysicsEngine::BulletPhysicsEngine(ecs3::World* world) : m_ecs(world), m_config(), m_world(nullptr) {
		m_config.broadphase = new btDbvtBroadphase();
		m_config.collisionConfiguration = new btDefaultCollisionConfiguration();
		m_config.dispatcher = new btCollisionDispatcher(m_config.collisionConfiguration);
		m_config.solver = new btSequentialImpulseConstraintSolver();

		m_world = new btDiscreteDynamicsWorld(
			m_config.dispatcher,
			m_config.broadphase,
			m_config.solver,
			m_config.collisionConfiguration);
	}

	void BulletPhysicsEngine::step() {
		checkForPhys();
		m_world->stepSimulation(60.0f);
		syncToECS();
	}

	static void build_motation_state(math::Transform* myState, BulletBody* btState) {
		auto myPos = myState->origin();
		btVector3 position{myPos.x, myPos.y, myPos.z};

		auto myRot = myState->euler();
		btQuaternion rotation;
		rotation.setEulerZYX(myRot.x, myRot.y, myRot.z);

		btTransform transform;
		transform.setIdentity();
		transform.setRotation(rotation);
		transform.setOrigin(position);

		btState->motion = new btDefaultMotionState(transform);
	}

	inline btCollisionShape* shapeToBullet(const phys::RigidBody* fgglBody) {
		if ( fgglBody->shape == nullptr ) {
			// they forgot to put a shape, we'll assume a unit sphere to avoid crashes...
			return new btSphereShape(0.5f);
		}

		switch (fgglBody->shape->type) {
			case phys::ShapeType::BOX: {
				auto extents = ((phys::Box*)fgglBody->shape)->extents;
				return new btBoxShape({extents.x, extents.y, extents.z});
			}
			case phys::ShapeType::SPHERE: {
				float radius = ((phys::Sphere*)fgglBody->shape)->radius;
				return new btSphereShape(radius);
			}
			default:
				return new btSphereShape(0.5f);
		}
	}

	void BulletPhysicsEngine::checkForPhys() {
		// FIXME without reactive-based approaches this is very slow
		auto entsWantPhys = m_ecs->findMatching<phys::RigidBody>();
		for (auto ent : entsWantPhys) {
			auto* btComp = m_ecs->get<BulletBody>(ent);
			if ( btComp != nullptr ) {
				// they are already in the simluation
				continue;
			}

			btComp = m_ecs->add<BulletBody>(ent);
			const auto* fgBody = m_ecs->get<phys::RigidBody>(ent);

			// setup the starting motion state
			auto* transform = m_ecs->get<math::Transform>(ent);
			build_motation_state(transform, btComp);

			btCollisionShape* colShape = shapeToBullet(fgBody);
			btVector3 localInt(0, 0, 0);
			if ( !fgBody->isStatic() ) {
				colShape->calculateLocalInertia(fgBody->mass, localInt);
			}

			// setup the construction information
			btRigidBody::btRigidBodyConstructionInfo bodyCI(
				fgBody->mass,
				btComp->motion,
				colShape,
				localInt
			);
			btComp->body = new btRigidBody(bodyCI);

			if (!fgBody->isStatic()) {
				// FIXME this is for testing, needs to be removed
				btComp->body->setRestitution(0.5f);
				btComp->body->setRollingFriction(0.0f);
			}
			m_world->addRigidBody(btComp->body);

			if ( !fgBody->isStatic() ) {
				btComp->body->applyTorque({50.0f, 0.5f, 0.5f});
				btComp->body->applyImpulse({5.0f, 0.5f, 0.5f}, {0.0f, 0.0f, 0.0f});
			}
		}
	}

	void BulletPhysicsEngine::syncToECS() {
		auto physEnts = m_ecs->findMatching<BulletBody>();
		for (auto& ent : physEnts) {
			auto* transform = m_ecs->get<math::Transform>(ent);
			auto* physBody = m_ecs->get<BulletBody>(ent);

			btTransform bTransform;
			physBody->motion->getWorldTransform(bTransform);

			// set calculate and set ecs position
			auto& btPos = bTransform.getOrigin();
			math::vec3 pos{btPos.x(), btPos.y(), btPos.z()};
			transform->origin( pos );

			// set the rotation
			const auto& btRot = bTransform.getRotation();
			math::quat rot{btRot.x(), btRot.y(), btRot.z(), btRot.w()};
			transform->setRotation(rot);
		}
	}

	BulletPhysicsEngine::~BulletPhysicsEngine() {
		// clean up the rigid bodies
		auto entRB = m_ecs->findMatching<BulletBody>();
		for (auto& ent : entRB) {
			auto* bulletBody = m_ecs->get<BulletBody>(ent);
			delete bulletBody->motion;
			m_world->removeCollisionObject(bulletBody->body);
			delete bulletBody->body;
			m_ecs->remove<BulletBody>(ent);
		}

		// delete the world
		delete m_world;

		// cleanup the configuration object
		delete m_config.solver;
		delete m_config.broadphase;
		delete m_config.dispatcher;
		delete m_config.collisionConfiguration;
	}

} // namespace fggl::phys::bullet